/*
**  Copyright (c) 2007 Sendmail, Inc. and its suppliers.
**    All rights reserved.
*/

#ifdef _FFR_QUERY_CACHE

#ifndef lint
static char dkim_cache_c_id[] = "@(#)$Id: dkim-cache.c,v 1.5 2007/04/25 06:15:32 msk Exp $";
#endif /* !lint */

/* system includes */
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <pthread.h>

/* libsm includes */
#include <sm/string.h>

/* libdb includes */
#include <db.h>

/* libdkim includes */
#include "dkim.h"

/* limits, macros, etc. */
#define	BUFRSZ			1024
#define DB_MODE			(S_IRUSR|S_IWUSR)

#ifndef DB_NOTFOUND
# define DB_NOTFOUND		1
#endif /* ! DB_NOTFOUND */

/* local definitions needed for DNS queries */
#ifndef MAXHOSTNAMELEN
# define MAXHOSTNAMELEN		256
#endif /* ! MAXHOSTNAMELEN */

/* data types */
struct dkim_cache_entry
{
	int		cache_ttl;
	time_t		cache_when;
	char		cache_data[BUFRSZ + 1];
};

/* globals */
static pthread_mutex_t cache_lock;

/*
**  DKIM_CACHE_INIT -- initialize an on-disk cache of entries
**
**  Parameters:
**  	err -- error code (returned)
**
**  Return value:
**  	A DB handle referring to the cache, or NULL on error.
*/

DB *
dkim_cache_init(int *err)
{
	int status = 0;
	DB *cache = NULL;

	(void) pthread_mutex_init(&cache_lock, NULL);

#if DB_VERSION_MAJOR > 2
	status = db_create(&cache, NULL, 0);
	if (status == 0)
	{
# if DB_VERSION_MAJOR > 3
		status = cache->open(cache, NULL, NULL, NULL, DB_HASH,
		                     (DB_CREATE|DB_THREAD), DB_MODE);
# else /* DB_VERSION_MAJOR > 3 */
		status = cache->open(cache, NULL, NULL, DB_HASH,
		                     (DB_CREATE|DB_THREAD), DB_MODE);
# endif /* DB_VERSION_MAJOR > 3 */
	}
#elif DB_VERSION_MAJOR == 2
	status = db_open(NULL, DB_HASH, (DB_CREATE|DB_THREAD), DB_MODE,
	                 NULL, NULL, &cache);
#else /* DB_VERSION_MAJOR < 2 */
	cache = dbopen(NULL, (O_CREAT|O_RDWR), DB_MODE, DB_HASH, NULL);
	if (cache == NULL)
		status = errno;
#endif /* DB_VERSION_MAJOR */

	if (status != 0)
	{
		if (err != NULL)
			*err = status;

		return NULL;
	}

#ifdef DEBUG
	printf("libdkim cache initialized\n");
#endif /* DEBUG */

	return cache;
}

/*
**  DKIM_CACHE_QUERY -- query an on-disk cache of entries
**
**  Parameters:
**  	db -- DB handle referring to the cache
**  	str -- key to query
**  	ttl -- time-to-live; ignore any record older than this; if 0, apply
**  	       the TTL in the record
**  	buf -- buffer into which to write any cached data found
**  	buflen -- number of bytes at "buffer" (returned); caller should set
**  	          this to the maximum space available and use the returned
**  	          value as the length of the data returned
**  	err -- error code (returned)
**
**  Return value:
**  	-1 -- error; caller should check "err"
**  	0 -- no error; record found and data returned
**  	1 -- no data found or data has expired
*/

int
dkim_cache_query(DB *db, char *str, int ttl, char *buf, size_t *buflen,
                 int *err)
{
	int status;
	time_t now;
	DBT q;
	DBT d;
	struct dkim_cache_entry ce;

	assert(db != NULL);
	assert(str != NULL);
	assert(buf != NULL);
	assert(err != NULL);

	memset(&q, '\0', sizeof q);
	memset(&d, '\0', sizeof d);

	q.data = str;
	q.size = strlen(q.data);

#ifdef DEBUG
	printf("libdkim cache query for `%s'\n", str);
#endif /* DEBUG */

#if DB_VERSION_MAJOR > 2
	d.flags = DB_DBT_USERMEM;
	d.data = (void *) &ce;
	d.size = sizeof ce;
#endif /* DB_VERSION_MAJOR > 2 */

	(void) time(&now);

	(void) pthread_mutex_lock(&cache_lock);

#if DB_VERSION_MAJOR < 2
	status = db->get(db, &q, &d, 0);
#else /* DB_VERSION_MAJOR < 2 */
	status = db->get(db, NULL, &q, &d, 0);
#endif /* DB_VERSION_MAJOR */

	(void) pthread_mutex_unlock(&cache_lock);

	if (status == 0)
	{
#if DB_VERSION_MAJOR < 2
		memset(&ce, '\0', sizeof ce);
		memcpy(&ce, d.data, MIN(sizeof ce, d.size));
#endif /* DB_VERSION_MAJOR < 2 */
		if (ttl != 0)
			ce.cache_ttl = ttl;
		if (ce.cache_when + ce.cache_ttl < now)
		{
#ifdef DEBUG
			printf("libdkim cache query for `%s' expired\n", str);
#endif /* DEBUG */
			return 1;
		}
		sm_strlcpy(buf, ce.cache_data, *buflen);
		*buflen = strlen(ce.cache_data);
#ifdef DEBUG
		printf("libdkim cache query for `%s' found: `%s'\n", str, buf);
#endif /* DEBUG */
		return 0;
	}
	else if (status != DB_NOTFOUND)
	{
		*err = status;
#ifdef DEBUG
		printf("libdkim cache query for `%s' error\n");
#endif /* DEBUG */
		return -1;
	}
	else
	{
#ifdef DEBUG
		printf("libdkim cache query for `%s' not found\n", str);
#endif /* DEBUG */
		return 1;
	}
}

/*
**  DKIM_CACHE_INSERT -- insert data into an on-disk cache of entries
**
**  Parameters:
**  	db -- DB handle referring to the cache
**  	str -- key to insert
**  	data -- data to insert
**  	ttl -- time-to-live
**  	err -- error code (returned)
**
**  Return value:
**  	-1 -- error; caller should check "err"
**  	0 -- cache updated
*/

int
dkim_cache_insert(DB *db, char *str, char *data, int ttl, int *err)
{
	int status;
	time_t now;
	DBT q;
	DBT d;
	struct dkim_cache_entry ce;

	assert(db != NULL);
	assert(str != NULL);
	assert(data != NULL);
	assert(err != NULL);

	(void) time(&now);

	memset(&q, '\0', sizeof q);
	memset(&d, '\0', sizeof d);

	q.data = str;
	q.size = strlen(str);

	d.data = (void *) &ce;
	d.size = sizeof ce;

	ce.cache_when = now;
	ce.cache_ttl = ttl;
	sm_strlcpy(ce.cache_data, data, sizeof ce.cache_data);

	(void) pthread_mutex_lock(&cache_lock);

#if DB_VERSION_MAJOR < 2
	status = db->put(db, &q, &d, 0);
#else /* DB_VERSION_MAJOR < 2 */
	status = db->put(db, NULL, &q, &d, 0);
#endif /* DB_VERSION_MAJOR */

	(void) pthread_mutex_unlock(&cache_lock);

	if (status == 0)
	{
#ifdef DEBUG
		printf("libdkim cache insert for `%s': `%s'\n", str, data);
#endif /* DEBUG */
		return 0;
	}
	else
	{
		*err = status;
#ifdef DEBUG
		printf("libdkim cache insert for `%s' error\n");
#endif /* DEBUG */
		return -1;
	}
}

/*
**  DKIM_CACHE_EXPIRE -- expire records in an on-disk cache of entries
**
**  Parameters:
**  	db -- DB handle referring to the cache
**  	ttl -- time-to-live; delete any record older than this; if 0, apply
**  	       the TTL in the record
**  	err -- error code (returned)
**
**  Return value:
**  	-1 -- error; caller should check "err"
**  	otherwise -- count of deleted records
*/

int
dkim_cache_expire(DB *db, int ttl, int *err)
{
	bool done = FALSE;
#if DB_VERSION_MAJOR < 2
	bool first = TRUE;
#endif /* DB_VERSION_MAJOR < 2 */
	bool delete;
	int deleted = 0;
	int status;
	time_t now;
#if DB_VERSION_MAJOR > 2
	DBC cursor;
#endif /* DB_VERSION_MAJOR > 2 */
	DBT q;
	DBT d;
	char name[MAXHOSTNAMELEN + 1];
	struct dkim_cache_entry ce;

	assert(db != NULL);
	assert(err != NULL);

	memset(&q, '\0', sizeof q);
	memset(&d, '\0', sizeof d);

	(void) time(&now);

	(void) pthread_mutex_lock(&cache_lock);

#if DB_VERSION_MAJOR > 2
	status = db->cursor(db, NULL, &dbc, 0);
	if (status != 0)
	{
		*err = status;
		return -1;
	}
#endif /* DB_VERSION_MAJOR > 2 */

	for (;;)
	{
		memset(name, '\0', sizeof name);
		memset(&ce, '\0', sizeof ce);

#if DB_VERSION_MAJOR > 2
		q.data = name;
		q.flags = DB_DBT_USERMEM;
		q.size = sizeof name;
#endif /* DB_VERSION_MAJOR > 2 */

#if DB_VERSION_MAJOR > 2
		d.data = (void *) &ce;
		d.flags = DB_DBT_USERMEM;
		d.size = sizeof ce;
#endif /* DB_VERSION_MAJOR > 2 */

#if DB_VERSION_MAJOR > 2
		status = dbc->c_get(dbc, &q, &d, DB_NEXT);
		if (status == DB_NOTFOUND)
		{
			break;
		}
		else if (status != 0)
		{
			*err = status;
			break;
		}
#else /* DB_VERSION_MAJOR > 2 */
		status = db->seq(db, &q, &d, first ? R_FIRST : R_NEXT);
		if (status == DB_NOTFOUND)
		{
			break;
		}
		else if (status != 0)
		{
			*err = status;
			break;
		}

		first = FALSE;

		memcpy(name, q.data, MIN(sizeof name, q.size));
		memcpy((void *) &ce, d.data, MIN(sizeof ce, d.size));
#endif /* DB_VERSION_MAJOR > 2 */

		delete = FALSE;
		if (ttl == 0)
		{
			if (ce.cache_when + ce.cache_ttl < now)
				delete = TRUE;
		}
		else
		{
			if (ce.cache_when + ttl < now)
				delete = TRUE;
		}

		if (delete)
		{
#if DB_VERSION_MAJOR > 2
			status = dbc->c_del(dbc, 0);
#else /* DB_VERSION_MAJOR > 2 */
			status = db->del(db, &q, R_CURSOR);
#endif /* DB_VERSION_MAJOR > 2 */
			if (status != 0)
			{
				*err = status;
				deleted = -1;
				break;
			}

			deleted++;
		}
	}

#if DB_VERSION_MAJOR > 2
	(void) dbc->close(dbc);
#endif /* DB_VERSION_MAJOR > 2 */

	(void) pthread_mutex_unlock(&cache_lock);

	return deleted;
}

/*
**  DKIM_CACHE_CLOSE -- close a cache database
**
**  Parameters:
**  	db -- cache DB handle
**
**  Return value:
**  	None.
*/

void
dkim_cache_close(DB *db)
{
	assert(db != NULL);

	(void) db->close(db);
}

#endif /* _FFR_QUERY_CACHE */
